@@ -28,12 +28,12 @@ class X::ParameterMismatch is Exception {
     }
 }
 
-# RAKUDO: rename to X::Property::NotFound once [RT #126827] has been fixed
-class X::PropertyNotFound is Exception {
+class X::Property::NotFound is Exception {
     has $.propname;
+    has $.type;
 
     method message {
-        "Property '$.propname' not found"
+        "Property '$.propname' not found on object of type $.type"
     }
 }
 
@@ -41,24 +41,35 @@ class X::Associativity::Conflict is Exception {
     method message { "The operator already has a defined associativity" }
 }
 
+class X::Regex::InvalidMatchType is Exception {
+    method message { "A regex can only match strings" }
+}
+
+class X::_007::RuntimeException is Exception {
+    has $.msg;
+
+    method message {
+        $.msg.Str;
+    }
+}
+
 sub aname($attr) { $attr.name.substr(2) }
 sub avalue($attr, $obj) { $attr.get_value($obj) }
 
 role Q {
     method Str {
-        my @attrs = self.attributes;
-        if @attrs == 1 {
-            return "{self.^name} { avalue(@attrs[0], self).quoted-Str }";
-        }
-        sub keyvalue($attr) { aname($attr) ~ ": " ~ avalue($attr, self).quoted-Str }
-        my $contents = @attrs.map(&keyvalue).join(",\n").indent(4);
-        return "{self.^name} \{\n$contents\n\}";
+        my %*stringification-seen;
+        Helper::Str(self);
     }
 
     method quoted-Str {
         self.Str
     }
 
+    method truthy {
+        True
+    }
+
     method attributes {
         sub find($aname) { self.^attributes.first({ $aname eq aname($_) }) }
 
@@ -78,7 +89,13 @@ role Q::Literal does Q::Term {
 }
 
 class Q::Literal::None does Q::Literal {
-    method eval($) { Val::None.new }
+    method eval($) { NONE }
+}
+
+class Q::Literal::Bool does Q::Literal {
+    has Val::Bool $.value;
+
+    method eval($) { $.value }
 }
 
 class Q::Literal::Int does Q::Literal {
@@ -95,15 +112,24 @@ class Q::Literal::Str does Q::Literal {
 
 class Q::Identifier does Q::Term {
     has Val::Str $.name;
-    has $.frame = Val::None.new;
+    has $.frame = NONE;
 
     method attribute-order { <name> }
 
     method eval($runtime) {
-        return $runtime.get-var(
-            $.name.value,
-            $.frame ~~ Val::None ?? $runtime.current-frame !! $.frame
-        );
+        return $runtime.get-var($.name.value, $.frame);
+    }
+
+    method put-value($value, $runtime) {
+        $runtime.put-var(self, $value);
+    }
+}
+
+class Q::Term::Regex does Q::Term {
+    has Val::Str $.contents;
+
+    method eval($runtime) {
+        Val::Regex.new(:$.contents);
     }
 }
 
@@ -111,7 +137,7 @@ class Q::Term::Array does Q::Term {
     has Val::Array $.elements;
 
     method eval($runtime) {
-        Val::Array.new(:elements($.elements.elements».eval($runtime)));
+        Val::Array.new(:elements($.elements.elements.map(*.eval($runtime))));
     }
 }
 
@@ -120,7 +146,7 @@ class Q::Term::Object does Q::Term {
     has $.propertylist;
 
     method eval($runtime) {
-        return $runtime.get-var($.type.name.value).create(
+        return $runtime.get-var($.type.name.value, $.type.frame).create(
             $.propertylist.properties.elements.map({.key.value => .value.eval($runtime)})
         );
     }
@@ -152,10 +178,6 @@ class Q::TraitList does Q {
     has Val::Array $.traits = Val::Array.new;
 
     method attribute-order { <traits> }
-
-    method interpolate($runtime) {
-        self.new(:traits(Val::Array.new(:elements($.traits.elements».interpolate($runtime)))));
-    }
 }
 
 class Q::Term::Sub does Q::Term does Q::Declaration {
@@ -166,13 +188,14 @@ class Q::Term::Sub does Q::Term does Q::Declaration {
     method attribute-order { <identifier traitlist block> }
 
     method eval($runtime) {
-        my $name = $.identifier ~~ Val::None
-            ?? "(anon)"
-            !! $.identifier.name.value;
+        my $name = $.identifier ~~ Val::NoneType
+            ?? Val::Str.new(:value(""))
+            !! $.identifier.name;
         return Val::Sub.new(
             :$name,
             :parameterlist($.block.parameterlist),
             :statementlist($.block.statementlist),
+            :static-lexpad($.block.static-lexpad),
             :outer-frame($runtime.current-frame),
         );
     }
@@ -181,16 +204,16 @@ class Q::Term::Sub does Q::Term does Q::Declaration {
 class Q::Block does Q {
     has $.parameterlist;
     has $.statementlist;
-    has %.static-lexpad;
+    has Val::Object $.static-lexpad is rw = Val::Object.new;
 
     method attribute-order { <parameterlist statementlist> }
 
-    method eval($runtime) {
-        my $outer-frame = $runtime.current-frame;
+    method reify($runtime) {
+        my Val::Object $outer-frame = $runtime.current-frame;
         Val::Block.new(
             :$.parameterlist,
             :$.statementlist,
-            :%.static-lexpad,
+            :$.static-lexpad,
             :$outer-frame
         );
     }
@@ -209,6 +232,10 @@ class Q::Prefix does Q::Expr {
     }
 }
 
+class Q::Prefix::Str is Q::Prefix {}
+
+class Q::Prefix::Plus is Q::Prefix {}
+
 class Q::Prefix::Minus is Q::Prefix {}
 
 class Q::Prefix::Not is Q::Prefix {}
@@ -236,6 +263,10 @@ class Q::Infix::Subtraction is Q::Infix {}
 
 class Q::Infix::Multiplication is Q::Infix {}
 
+class Q::Infix::Modulo is Q::Infix {}
+
+class Q::Infix::Divisibility is Q::Infix {}
+
 class Q::Infix::Concat is Q::Infix {}
 
 class Q::Infix::Replicate is Q::Infix {}
@@ -246,10 +277,8 @@ class Q::Infix::Cons is Q::Infix {}
 
 class Q::Infix::Assignment is Q::Infix {
     method eval($runtime) {
-        die "Needs to be an identifier on the left"     # XXX: Turn this into an X::
-            unless $.lhs ~~ Q::Identifier;
         my $value = $.rhs.eval($runtime);
-        $runtime.put-var($.lhs.name.value, $value);
+        $.lhs.put-value($value, $runtime);
         return $value;
     }
 }
@@ -275,6 +304,15 @@ class Q::Infix::Or is Q::Infix {
     }
 }
 
+class Q::Infix::DefinedOr is Q::Infix {
+    method eval($runtime) {
+        my $l = $.lhs.eval($runtime);
+        return $l !~~ Val::NoneType
+            ?? $l
+            !! $.rhs.eval($runtime);
+    }
+}
+
 class Q::Infix::And is Q::Infix {
     method eval($runtime) {
         my $l = $.lhs.eval($runtime);
@@ -284,7 +322,9 @@ class Q::Infix::And is Q::Infix {
     }
 }
 
-class Q::Infix::TypeEq is Q::Infix {}
+class Q::Infix::TypeMatch is Q::Infix {}
+
+class Q::Infix::TypeNonMatch is Q::Infix {}
 
 class Q::Postfix does Q::Expr {
     has $.identifier;
@@ -316,7 +356,7 @@ class Q::Postfix::Index is Q::Postfix {
                     if $index.value < 0;
                 return .elements[$index.value];
             }
-            when Val::Object | Q {
+            when Val::Object | Val::Block | Q {
                 my $property = $.index.eval($runtime);
                 die X::Subscript::NonString.new
                     if $property !~~ Val::Str;
@@ -326,6 +366,29 @@ class Q::Postfix::Index is Q::Postfix {
             die X::TypeCheck.new(:operation<indexing>, :got($_), :expected(Val::Array));
         }
     }
+
+    method put-value($value, $runtime) {
+        given $.operand.eval($runtime) {
+            when Val::Array {
+                my $index = $.index.eval($runtime);
+                die X::Subscript::NonInteger.new
+                    if $index !~~ Val::Int;
+                die X::Subscript::TooLarge.new(:value($index.value), :length(+.elements))
+                    if $index.value >= .elements;
+                die X::Subscript::Negative.new(:$index, :type([]))
+                    if $index.value < 0;
+                .elements[$index.value] = $value;
+            }
+            when Val::Object | Q {
+                my $property = $.index.eval($runtime);
+                die X::Subscript::NonString.new
+                    if $property !~~ Val::Str;
+                my $propname = $property.value;
+                $runtime.put-property($_, $propname, $value);
+            }
+            die X::TypeCheck.new(:operation<indexing>, :got($_), :expected(Val::Array));
+        }
+    }
 }
 
 class Q::Postfix::Call is Q::Postfix {
@@ -339,7 +402,7 @@ class Q::Postfix::Call is Q::Postfix {
             if $c ~~ Val::Macro;
         die "Trying to invoke a {$c.^name.subst(/^'Val::'/, '')}" # XXX: make this into an X::
             unless $c ~~ Val::Block;
-        my @arguments = $.argumentlist.arguments.elements».eval($runtime);
+        my @arguments = $.argumentlist.arguments.elements.map(*.eval($runtime));
         return $runtime.call($c, @arguments);
     }
 }
@@ -354,6 +417,16 @@ class Q::Postfix::Property is Q::Postfix {
         my $propname = $.property.name.value;
         $runtime.property($obj, $propname);
     }
+
+    method put-value($value, $runtime) {
+        given $.operand.eval($runtime) {
+            when Val::Object | Q {
+                my $propname = $.property.name.value;
+                $runtime.put-property($_, $propname, $value);
+            }
+            die "We don't handle this case yet"; # XXX: think more about this case
+        }
+    }
 }
 
 class Q::Unquote does Q {
@@ -374,8 +447,11 @@ class Q::Unquote::Infix is Q::Unquote {
 }
 
 class Q::Term::Quasi does Q::Term {
+    has $.qtype;
     has $.contents;
 
+    method attribute-order { <qtype contents> }
+
     method eval($runtime) {
         sub interpolate($thing) {
             return $thing.new(:elements($thing.elements.map(&interpolate)))
@@ -417,24 +493,24 @@ class Q::Term::Quasi does Q::Term {
             $thing.new(|%attributes);
         }
 
-        if $.contents ~~ Q::Unquote {   # special exception: `quasi @ Q::Unquote`
+        if $.qtype.value eq "Q::Unquote" && $.contents ~~ Q::Unquote {
             return $.contents;
         }
         return interpolate($.contents);
     }
 }
 
-class Q::ParameterList does Q {
-    # RAKUDO: Can simplify this to `.=` once [RT #126975] is fixed
-    has Val::Array $.parameters = Val::Array.new;
-}
-
 class Q::Parameter does Q does Q::Declaration {
     has $.identifier;
 
     method is-assignable { True }
 }
 
+class Q::ParameterList does Q {
+    # RAKUDO: Can simplify this to `.=` once [RT #126975] is fixed
+    has Val::Array $.parameters = Val::Array.new;
+}
+
 class Q::ArgumentList does Q {
     # RAKUDO: Can simplify this to `.=` once [RT #126975] is fixed
     has Val::Array $.arguments = Val::Array.new;
@@ -445,7 +521,7 @@ role Q::Statement does Q {
 
 class Q::Statement::My does Q::Statement does Q::Declaration {
     has $.identifier;
-    has $.expr = Val::None.new;
+    has $.expr = NONE;
 
     method attribute-order { <identifier expr> }
 
@@ -453,9 +529,9 @@ class Q::Statement::My does Q::Statement does Q::Declaration {
 
     method run($runtime) {
         return
-            unless $.expr !~~ Val::None;
+            unless $.expr !~~ Val::NoneType;
         my $value = $.expr.eval($runtime);
-        $runtime.put-var($.identifier.name.value, $value);
+        $.identifier.put-value($value, $runtime);
     }
 }
 
@@ -481,33 +557,32 @@ class Q::Statement::Expr does Q::Statement {
 class Q::Statement::If does Q::Statement {
     has $.expr;
     has $.block;
-    has $.else = Val::None.new;
+    has $.else = NONE;
 
     method attribute-order { <expr block else> }
 
     method run($runtime) {
         my $expr = $.expr.eval($runtime);
         if $expr.truthy {
-            my $c = $.block.eval($runtime);
+            my $c = $.block.reify($runtime);
             $runtime.enter($c);
             my $paramcount = $c.parameterlist.elems;
             die X::ParameterMismatch.new(
                 :type("If statement"), :$paramcount, :argcount("0 or 1"))
                 if $paramcount > 1;
             for @($c.parameterlist.parameters.elements) Z $expr -> ($param, $arg) {
-                $runtime.declare-var($param.identifier.name.value, $arg);
+                $runtime.declare-var($param.identifier, $arg);
             }
             $.block.statementlist.run($runtime);
             $runtime.leave;
         }
         else {
             given $.else {
-                when Val::None { }
                 when Q::Statement::If {
                     $.else.run($runtime)
                 }
                 when Q::Block {
-                    my $c = $.else.eval($runtime);
+                    my $c = $.else.reify($runtime);
                     $runtime.enter($c);
                     $.else.statementlist.run($runtime);
                     $runtime.leave;
@@ -521,7 +596,7 @@ class Q::Statement::Block does Q::Statement {
     has $.block;
 
     method run($runtime) {
-        $runtime.enter($.block.eval($runtime));
+        $runtime.enter($.block.reify($runtime));
         $.block.statementlist.run($runtime);
         $runtime.leave;
     }
@@ -537,43 +612,23 @@ class Q::Statement::For does Q::Statement {
     method attribute-order { <expr block> }
 
     method run($runtime) {
-        multi split_elements(@array, 1) { return @array }
-        multi split_elements(@array, Int $n) {
-            my $list = @array.list;
-            my @split;
-
-            while True {
-                my @new = $list.splice(0, $n);
-                last unless @new;
-                @split.push: @new.item;
-            }
-
-            @split;
-        }
-
-        my $c = $.block.eval($runtime);
+        my $c = $.block.reify($runtime);
         my $count = $c.parameterlist.parameters.elements.elems;
+        die X::ParameterMismatch.new(
+            :type("For loop"), :paramcount($count), :argcount("0 or 1"))
+            if $count > 1;
 
         my $array = $.expr.eval($runtime);
         die X::TypeCheck.new(:operation("for loop"), :got($array), :expected(Val::Array))
             unless $array ~~ Val::Array;
 
-        if $count == 0 {
-            for $array.elements {
-                $runtime.enter($c);
-                $.block.statementlist.run($runtime);
-                $runtime.leave;
-            }
-        }
-        else {
-            for split_elements($array.elements, $count) -> $arg {
-                $runtime.enter($c);
-                for @($c.parameterlist.parameters.elements) Z $arg.list -> ($param, $real_arg) {
-                    $runtime.declare-var($param.identifier.name.value, $real_arg);
-                }
-                $.block.statementlist.run($runtime);
-                $runtime.leave;
+        for $array.elements -> $arg {
+            $runtime.enter($c);
+            if $count == 1 {
+                $runtime.declare-var($c.parameterlist.parameters.elements[0].identifier, $arg.list[0]);
             }
+            $.block.statementlist.run($runtime);
+            $runtime.leave;
         }
     }
 }
@@ -586,14 +641,14 @@ class Q::Statement::While does Q::Statement {
 
     method run($runtime) {
         while (my $expr = $.expr.eval($runtime)).truthy {
-            my $c = $.block.eval($runtime);
+            my $c = $.block.reify($runtime);
             $runtime.enter($c);
             my $paramcount = $c.parameterlist.parameters.elements.elems;
             die X::ParameterMismatch.new(
                 :type("While loop"), :$paramcount, :argcount("0 or 1"))
                 if $paramcount > 1;
             for @($c.parameterlist.parameters.elements) Z $expr -> ($param, $arg) {
-                $runtime.declare-var($param.identifier.name.value, $arg);
+                $runtime.declare-var($param.identifier, $arg);
             }
             $.block.statementlist.run($runtime);
             $runtime.leave;
@@ -602,19 +657,33 @@ class Q::Statement::While does Q::Statement {
 }
 
 class Q::Statement::Return does Q::Statement {
-    has $.expr = Val::None.new;
+    has $.expr = NONE;
 
     method run($runtime) {
-        my $value = $.expr ~~ Val::None ?? $.expr !! $.expr.eval($runtime);
+        my $value = $.expr ~~ Val::NoneType ?? $.expr !! $.expr.eval($runtime);
         my $frame = $runtime.get-var("--RETURN-TO--");
         die X::Control::Return.new(:$value, :$frame);
     }
 }
 
+class Q::Statement::Throw does Q::Statement {
+    has $.expr = NONE;
+
+    method run($runtime) {
+        my $value = $.expr ~~ Val::NoneType
+            ?? Val::Exception.new(:message(Val::Str.new(:value("Died"))))
+            !! $.expr.eval($runtime);
+        die X::TypeCheck.new(:got($value), :excpected(Val::Exception))
+            if $value !~~ Val::Exception;
+
+        die X::_007::RuntimeException.new(:msg($value.message.value));
+    }
+}
+
 class Q::Statement::Sub does Q::Statement does Q::Declaration {
     has $.identifier;
     has $.traitlist = Q::TraitList.new;
-    has $.block;
+    has Q::Block $.block;
 
     method attribute-order { <identifier traitlist block> }
 
@@ -641,6 +710,14 @@ class Q::Statement::BEGIN does Q::Statement {
     }
 }
 
+class Q::Statement::Class does Q::Statement does Q::Declaration {
+    has $.block;
+
+    method run($runtime) {
+        # a class block does not run at runtime
+    }
+}
+
 class Q::StatementList does Q {
     # RAKUDO: Can simplify this to `.=` once [RT #126975] is fixed
     has Val::Array $.statements = Val::Array.new;
@@ -651,3 +728,12 @@ class Q::StatementList does Q {
         }
     }
 }
+
+class Q::Expr::StatementListAdapter does Q::Expr {
+    has $.statementlist;
+
+    method eval($runtime) {
+        $.statementlist.run($runtime);
+        return NONE;
+    }
+}
